/* 

 * OpenGL.c

 *

 * Part of the Plasma Game Server

 * written/copyright Ryan Haksi, 1998

 * 

 * If you use this file as part of a publically released project, please 

 * grant me the courtesy of a mention in the credits section somewhere.

 * Other than that opengl.h and opengl.c are yours to do with what you wish

 *

 * Current email address(es): cryogen@infoserve.net

 * Current web page address: http://members.home.com/borealis/ryan.html

 * 

 * If you add functions to this file, please email me the updated copy so that

 * I can maintain the most up to date copy.



/*

 * A special thanks goes out to Dan Mintz ( brink@silcon.com ) for providing the full opengl function list

 * as well as the the ARB multitexture and EXT compiled array stuff.

 */

/* 

 * More thanx goes out to "Jorrit Rouwe" <jorrit@three14.demon.nl> for the macro idea

 * to shadow the wgl calls.

 */



#define __PLASMA_OPENGL_C



#include <windows.h>

#include <stdlib.h>

#include <stdio.h>



#include "opengl.h"



/* Set to 1 to use support only minigl function calls, and locate minigl drivers */

#define MINIGL 0



/* set to 1 to autoload popular extensions - if you are using multiple contexts this wont work */

#define AUTOLOAD_EXTENSIONS 1



int       openglInit = 0;

HINSTANCE openglInst = NULL;

HINSTANCE openglGLUInst = NULL;

BOOL      openglBypassGDI = FALSE;

char      openglGLUDriver[128];



//////////////////////////////////////////////////////////////////////////



int OpenGLDetect3dfx(void) {

  typedef struct GlideHwConfigStruct GLIDEHWCONFIG;

  struct GlideHwConfigStruct {

    int   num_sst; 

    char  filler[4096]; // hope this is big enough

  };



  BOOL              (__stdcall* GLIDEFN_grSstQueryBoards) (GLIDEHWCONFIG* hwconfig);

  int               retValue = FALSE;

  HINSTANCE         hGlide = NULL;

  GLIDEHWCONFIG     glideHwConfig;



  // load glide

  hGlide = LoadLibrary("glide2x.dll");

  if (!hGlide) return FALSE;



  GLIDEFN_grSstQueryBoards = (void*) GetProcAddress(hGlide, "_grSstQueryBoards@4");



  if (GLIDEFN_grSstQueryBoards) {

    GLIDEFN_grSstQueryBoards(&glideHwConfig);

    if (glideHwConfig.num_sst > 0) 

      retValue = TRUE;

  } 

  FreeLibrary(hGlide);

  return retValue;

}



//////////////////////////////////////////////////////////////////////////



void OpenGLFindDriver(char *openglLib) {

  openglLib[0]=0;



  /* attempt to autodetermine opengl driver */

  /* Voodoo Rush, Voodoo1, Voodoo2 */

  if (!*openglLib) {

    HINSTANCE found = NULL;



    if ( OpenGLDetect3dfx() ) {

#if (MINIGL)

      strcpy(openglLib, "3dfxgl.dll");

      found = LoadLibrary(openglLib);

      if (!found) {

        strcpy(openglLib, "3dfxvgl.dll");

        found = LoadLibrary(openglLib);

      }

#else

      strcpy(openglLib, "3dfxvgl.dll");

      found = LoadLibrary(openglLib);

#endif

      if (found) {

        FreeLibrary(found);

        return;

      }

    }

  }



#if (MINIGL)

  /* PowerVR? Will this work */

  if (!*openglLib) {

    HINSTANCE found = NULL;

    found = LoadLibrary("pvrhal32.dll");

    if (found) {

      FreeLibrary(found);

      strcpy(openglLib, "pvrgl.dll");

      found = LoadLibrary(openglLib);

      if (found) {

        FreeLibrary(found);

        return;

      }

    }

  }

#endif



  /* Check for SGI opengl better s/w performance than MS */

  if (!*openglLib) {

    HINSTANCE found = NULL;

    found = LoadLibrary("opengl.dll");

    if (found) {

      FreeLibrary(found);

      strcpy(openglLib, "opengl.dll");

    }

  }



  /* Fallback to system opengl MCD/ICD driver */

  if (!*openglLib) {

    strcpy(openglLib, "opengl32.dll");

  }

}



int OpenGLInit(char *lib) {

  openglInit+=1;

  if (openglInit>1) return 0;



  /* load dll */ 

  openglInst = LoadLibrary(lib); 

  if (openglInst == NULL) { 

    MessageBox(GetFocus(), "Could not open OpenGL Library", "Error", MB_OK); 

    openglInst = NULL; 

    return -666; 

  }



#if (!MINIGL)

  /* It would be better if this supported full paths */

  if (!strcmp(lib, "opengl.dll"))   openglGLUInst = LoadLibrary("glu.dll"); 

  if (!strcmp(lib, "opengl32.dll")) openglGLUInst = LoadLibrary("glu32.dll"); 

  if (!strcmp(lib, "3dfxvgl.dll"))  openglGLUInst = LoadLibrary("3dfxvglu.dll"); 

#endif



#define OPENGL_LOAD_GL_FN(f)    if (!(fn##f = (void*)GetProcAddress(openglInst, #f))) return -1

#define OPENGL_LOAD_GLU_FN(f)   if (!(fn##f = (void*)GetProcAddress(openglGLUInst, #f))) return -1

#define OPENGL_LOAD_EXT_FN(f)   fn##f = (void *)OpenGLGetProcAddress(#f)



  /* load all the function pointers */

#if (MINIGL)

  OPENGL_LOAD_GL_FN( glAlphaFunc     );

  OPENGL_LOAD_GL_FN( glBegin         );

  OPENGL_LOAD_GL_FN( glBindTexture   );

  OPENGL_LOAD_GL_FN( glBlendFunc     );

  OPENGL_LOAD_GL_FN( glClear         );

  OPENGL_LOAD_GL_FN( glClearColor    );

  OPENGL_LOAD_GL_FN( glClearDepth    );

  OPENGL_LOAD_GL_FN( glColor3f       );

  OPENGL_LOAD_GL_FN( glColor3fv      );

  OPENGL_LOAD_GL_FN( glColor4f       );

  OPENGL_LOAD_GL_FN( glColor4fv      );

  OPENGL_LOAD_GL_FN( glCullFace      );

  OPENGL_LOAD_GL_FN( glDepthFunc     );

  OPENGL_LOAD_GL_FN( glDisable       );

  OPENGL_LOAD_GL_FN( glEnable        );

  OPENGL_LOAD_GL_FN( glEnd           );

  OPENGL_LOAD_GL_FN( glFinish        );

  OPENGL_LOAD_GL_FN( glFrontFace     );

  OPENGL_LOAD_GL_FN( glFrustum       );

  OPENGL_LOAD_GL_FN( glGetError      );

  OPENGL_LOAD_GL_FN( glGetFloatv     );

  OPENGL_LOAD_GL_FN( glGetIntegerv   );

  OPENGL_LOAD_GL_FN( glGetString     );

  OPENGL_LOAD_GL_FN( glHint          );

  OPENGL_LOAD_GL_FN( glIsEnabled     );

  OPENGL_LOAD_GL_FN( glLoadIdentity  );

  OPENGL_LOAD_GL_FN( glLoadMatrixf   );

  OPENGL_LOAD_GL_FN( glMatrixMode    );

  OPENGL_LOAD_GL_FN( glMultMatrixf   );

  OPENGL_LOAD_GL_FN( glOrtho         );

  OPENGL_LOAD_GL_FN( glPixelStorei   );

  OPENGL_LOAD_GL_FN( glPolygonOffset );

  OPENGL_LOAD_GL_FN( glPopMatrix     );

  OPENGL_LOAD_GL_FN( glPushMatrix    );

  OPENGL_LOAD_GL_FN( glRotatef       );

  OPENGL_LOAD_GL_FN( glScalef        );

  OPENGL_LOAD_GL_FN( glShadeModel    );

  OPENGL_LOAD_GL_FN( glTexCoord2f    );

  OPENGL_LOAD_GL_FN( glTexCoord2fv   );

  OPENGL_LOAD_GL_FN( glTexCoord3f    );

  OPENGL_LOAD_GL_FN( glTexCoord3fv   );

  OPENGL_LOAD_GL_FN( glTexEnvf       );

  OPENGL_LOAD_GL_FN( glTexImage2D    );

  OPENGL_LOAD_GL_FN( glTexParameterf );

  OPENGL_LOAD_GL_FN( glTranslatef    );

  OPENGL_LOAD_GL_FN( glVertex2f      );

  OPENGL_LOAD_GL_FN( glVertex3f      );

  OPENGL_LOAD_GL_FN( glVertex3fv     );

  OPENGL_LOAD_GL_FN( glVertex4f      );

  OPENGL_LOAD_GL_FN( glViewport      );

#else

  OPENGL_LOAD_GL_FN( glAccum                  );

  OPENGL_LOAD_GL_FN( glAlphaFunc              );

  OPENGL_LOAD_GL_FN( glAreTexturesResident    );

  OPENGL_LOAD_GL_FN( glArrayElement           );

  OPENGL_LOAD_GL_FN( glBegin                  );

  OPENGL_LOAD_GL_FN( glBindTexture            );

  OPENGL_LOAD_GL_FN( glBitmap                 );

  OPENGL_LOAD_GL_FN( glBlendFunc              );

  OPENGL_LOAD_GL_FN( glCallList               );

  OPENGL_LOAD_GL_FN( glCallLists              );

  OPENGL_LOAD_GL_FN( glClear                  );

  OPENGL_LOAD_GL_FN( glClearAccum             );

  OPENGL_LOAD_GL_FN( glClearColor             );

  OPENGL_LOAD_GL_FN( glClearDepth             );

  OPENGL_LOAD_GL_FN( glClearIndex             );

  OPENGL_LOAD_GL_FN( glClearStencil           );

  OPENGL_LOAD_GL_FN( glClipPlane              );

  OPENGL_LOAD_GL_FN( glColor3b                );

  OPENGL_LOAD_GL_FN( glColor3bv               );

  OPENGL_LOAD_GL_FN( glColor3d                );

  OPENGL_LOAD_GL_FN( glColor3dv               );

  OPENGL_LOAD_GL_FN( glColor3f                );

  OPENGL_LOAD_GL_FN( glColor3fv               );

  OPENGL_LOAD_GL_FN( glColor3i                );

  OPENGL_LOAD_GL_FN( glColor3iv               );

  OPENGL_LOAD_GL_FN( glColor3s                );

  OPENGL_LOAD_GL_FN( glColor3sv               );

  OPENGL_LOAD_GL_FN( glColor3ub               );

  OPENGL_LOAD_GL_FN( glColor3ubv              );

  OPENGL_LOAD_GL_FN( glColor3ui               );

  OPENGL_LOAD_GL_FN( glColor3uiv              );

  OPENGL_LOAD_GL_FN( glColor3us               );

  OPENGL_LOAD_GL_FN( glColor3usv              );

  OPENGL_LOAD_GL_FN( glColor4b                );

  OPENGL_LOAD_GL_FN( glColor4bv               );

  OPENGL_LOAD_GL_FN( glColor4d                );

  OPENGL_LOAD_GL_FN( glColor4dv               );

  OPENGL_LOAD_GL_FN( glColor4f                );

  OPENGL_LOAD_GL_FN( glColor4fv               );

  OPENGL_LOAD_GL_FN( glColor4i                );

  OPENGL_LOAD_GL_FN( glColor4iv               );

  OPENGL_LOAD_GL_FN( glColor4s                );

  OPENGL_LOAD_GL_FN( glColor4sv               );

  OPENGL_LOAD_GL_FN( glColor4ub               );

  OPENGL_LOAD_GL_FN( glColor4ubv              );

  OPENGL_LOAD_GL_FN( glColor4ui               );

  OPENGL_LOAD_GL_FN( glColor4uiv              );

  OPENGL_LOAD_GL_FN( glColor4us               );

  OPENGL_LOAD_GL_FN( glColor4usv              );

  OPENGL_LOAD_GL_FN( glColorMask              );

  OPENGL_LOAD_GL_FN( glColorMaterial          );

  OPENGL_LOAD_GL_FN( glColorPointer           );

  OPENGL_LOAD_GL_FN( glCopyPixels             );

  OPENGL_LOAD_GL_FN( glCopyTexImage1D         );

  OPENGL_LOAD_GL_FN( glCopyTexImage2D         );

  OPENGL_LOAD_GL_FN( glCopyTexSubImage1D      );

  OPENGL_LOAD_GL_FN( glCopyTexSubImage2D      );

  OPENGL_LOAD_GL_FN( glCullFace               );

  OPENGL_LOAD_GL_FN( glDeleteLists            );

  OPENGL_LOAD_GL_FN( glDeleteTextures         );

  OPENGL_LOAD_GL_FN( glDepthFunc              );

  OPENGL_LOAD_GL_FN( glDepthMask              );

  OPENGL_LOAD_GL_FN( glDepthRange             );

  OPENGL_LOAD_GL_FN( glDisable                );

  OPENGL_LOAD_GL_FN( glDisableClientState     );

  OPENGL_LOAD_GL_FN( glDrawArrays             );

  OPENGL_LOAD_GL_FN( glDrawBuffer             );

  OPENGL_LOAD_GL_FN( glDrawElements           );

  OPENGL_LOAD_GL_FN( glDrawPixels             );

  OPENGL_LOAD_GL_FN( glEdgeFlag               );

  OPENGL_LOAD_GL_FN( glEdgeFlagPointer        );

  OPENGL_LOAD_GL_FN( glEdgeFlagv              );

  OPENGL_LOAD_GL_FN( glEnable                 );

  OPENGL_LOAD_GL_FN( glEnableClientState      );

  OPENGL_LOAD_GL_FN( glEnd                    );

  OPENGL_LOAD_GL_FN( glEndList                );

  OPENGL_LOAD_GL_FN( glEvalCoord1d            );

  OPENGL_LOAD_GL_FN( glEvalCoord1dv           );

  OPENGL_LOAD_GL_FN( glEvalCoord1f            );

  OPENGL_LOAD_GL_FN( glEvalCoord1fv           );

  OPENGL_LOAD_GL_FN( glEvalCoord2d            );

  OPENGL_LOAD_GL_FN( glEvalCoord2dv           );

  OPENGL_LOAD_GL_FN( glEvalCoord2f            );

  OPENGL_LOAD_GL_FN( glEvalCoord2fv           );

  OPENGL_LOAD_GL_FN( glEvalMesh1              );

  OPENGL_LOAD_GL_FN( glEvalMesh2              );

  OPENGL_LOAD_GL_FN( glEvalPoint1             );

  OPENGL_LOAD_GL_FN( glEvalPoint2             );

  OPENGL_LOAD_GL_FN( glFeedbackBuffer         );

  OPENGL_LOAD_GL_FN( glFinish                 );

  OPENGL_LOAD_GL_FN( glFlush                  );

  OPENGL_LOAD_GL_FN( glFogf                   );

  OPENGL_LOAD_GL_FN( glFogfv                  );

  OPENGL_LOAD_GL_FN( glFogi                   );

  OPENGL_LOAD_GL_FN( glFogiv                  );

  OPENGL_LOAD_GL_FN( glFrontFace              );

  OPENGL_LOAD_GL_FN( glFrustum                );

  OPENGL_LOAD_GL_FN( glGenLists               );

  OPENGL_LOAD_GL_FN( glGenTextures            );

  OPENGL_LOAD_GL_FN( glGetBooleanv            );

  OPENGL_LOAD_GL_FN( glGetClipPlane           );

  OPENGL_LOAD_GL_FN( glGetDoublev             );

  OPENGL_LOAD_GL_FN( glGetError               );

  OPENGL_LOAD_GL_FN( glGetFloatv              );

  OPENGL_LOAD_GL_FN( glGetIntegerv            );

  OPENGL_LOAD_GL_FN( glGetLightfv             );

  OPENGL_LOAD_GL_FN( glGetLightiv             );

  OPENGL_LOAD_GL_FN( glGetMapdv               );

  OPENGL_LOAD_GL_FN( glGetMapfv               );

  OPENGL_LOAD_GL_FN( glGetMapiv               );

  OPENGL_LOAD_GL_FN( glGetMaterialfv          );

  OPENGL_LOAD_GL_FN( glGetMaterialiv          );

  OPENGL_LOAD_GL_FN( glGetPixelMapfv          );

  OPENGL_LOAD_GL_FN( glGetPixelMapuiv         );

  OPENGL_LOAD_GL_FN( glGetPixelMapusv         );

  OPENGL_LOAD_GL_FN( glGetPointerv            );

  OPENGL_LOAD_GL_FN( glGetPolygonStipple      );

  OPENGL_LOAD_GL_FN( glGetString              );

  OPENGL_LOAD_GL_FN( glGetTexEnvfv            );

  OPENGL_LOAD_GL_FN( glGetTexEnviv            );

  OPENGL_LOAD_GL_FN( glGetTexGendv            );

  OPENGL_LOAD_GL_FN( glGetTexGenfv            );

  OPENGL_LOAD_GL_FN( glGetTexGeniv            );

  OPENGL_LOAD_GL_FN( glGetTexImage            );

  OPENGL_LOAD_GL_FN( glGetTexLevelParameterfv );

  OPENGL_LOAD_GL_FN( glGetTexLevelParameteriv );

  OPENGL_LOAD_GL_FN( glGetTexParameterfv      );

  OPENGL_LOAD_GL_FN( glGetTexParameteriv      );

  OPENGL_LOAD_GL_FN( glHint                   );

  OPENGL_LOAD_GL_FN( glIndexMask              );

  OPENGL_LOAD_GL_FN( glIndexPointer           );

  OPENGL_LOAD_GL_FN( glIndexd                 );

  OPENGL_LOAD_GL_FN( glIndexdv                );

  OPENGL_LOAD_GL_FN( glIndexf                 );

  OPENGL_LOAD_GL_FN( glIndexfv                );

  OPENGL_LOAD_GL_FN( glIndexi                 );

  OPENGL_LOAD_GL_FN( glIndexiv                );

  OPENGL_LOAD_GL_FN( glIndexs                 );

  OPENGL_LOAD_GL_FN( glIndexsv                );

  OPENGL_LOAD_GL_FN( glIndexub                );

  OPENGL_LOAD_GL_FN( glIndexubv               );

  OPENGL_LOAD_GL_FN( glInitNames              );

  OPENGL_LOAD_GL_FN( glInterleavedArrays      );

  OPENGL_LOAD_GL_FN( glIsEnabled              );

  OPENGL_LOAD_GL_FN( glIsList                 );

  OPENGL_LOAD_GL_FN( glIsTexture              );

  OPENGL_LOAD_GL_FN( glLightModelf            );

  OPENGL_LOAD_GL_FN( glLightModelfv           );

  OPENGL_LOAD_GL_FN( glLightModeli            );

  OPENGL_LOAD_GL_FN( glLightModeliv           );

  OPENGL_LOAD_GL_FN( glLightf                 );

  OPENGL_LOAD_GL_FN( glLightfv                );

  OPENGL_LOAD_GL_FN( glLighti                 );

  OPENGL_LOAD_GL_FN( glLightiv                );

  OPENGL_LOAD_GL_FN( glLineStipple            );

  OPENGL_LOAD_GL_FN( glLineWidth              );

  OPENGL_LOAD_GL_FN( glListBase               );

  OPENGL_LOAD_GL_FN( glLoadIdentity           );

  OPENGL_LOAD_GL_FN( glLoadMatrixd            );

  OPENGL_LOAD_GL_FN( glLoadMatrixf            );

  OPENGL_LOAD_GL_FN( glLoadName               );

  OPENGL_LOAD_GL_FN( glLogicOp                );

  OPENGL_LOAD_GL_FN( glMap1d                  );

  OPENGL_LOAD_GL_FN( glMap1f                  );

  OPENGL_LOAD_GL_FN( glMap2d                  );

  OPENGL_LOAD_GL_FN( glMap2f                  );

  OPENGL_LOAD_GL_FN( glMapGrid1d              );

  OPENGL_LOAD_GL_FN( glMapGrid1f              );

  OPENGL_LOAD_GL_FN( glMapGrid2d              );

  OPENGL_LOAD_GL_FN( glMapGrid2f              );

  OPENGL_LOAD_GL_FN( glMaterialf              );

  OPENGL_LOAD_GL_FN( glMaterialfv             );

  OPENGL_LOAD_GL_FN( glMateriali              );

  OPENGL_LOAD_GL_FN( glMaterialiv             );

  OPENGL_LOAD_GL_FN( glMatrixMode             );

  OPENGL_LOAD_GL_FN( glMultMatrixd            );

  OPENGL_LOAD_GL_FN( glMultMatrixf            );

  OPENGL_LOAD_GL_FN( glNewList                );

  OPENGL_LOAD_GL_FN( glNormal3b               );

  OPENGL_LOAD_GL_FN( glNormal3bv              );

  OPENGL_LOAD_GL_FN( glNormal3d               );

  OPENGL_LOAD_GL_FN( glNormal3dv              );

  OPENGL_LOAD_GL_FN( glNormal3f               );

  OPENGL_LOAD_GL_FN( glNormal3fv              );

  OPENGL_LOAD_GL_FN( glNormal3i               );

  OPENGL_LOAD_GL_FN( glNormal3iv              );

  OPENGL_LOAD_GL_FN( glNormal3s               );

  OPENGL_LOAD_GL_FN( glNormal3sv              );

  OPENGL_LOAD_GL_FN( glNormalPointer          );

  OPENGL_LOAD_GL_FN( glOrtho                  );

  OPENGL_LOAD_GL_FN( glPassThrough            );

  OPENGL_LOAD_GL_FN( glPixelMapfv             );

  OPENGL_LOAD_GL_FN( glPixelMapuiv            );

  OPENGL_LOAD_GL_FN( glPixelMapusv            );

  OPENGL_LOAD_GL_FN( glPixelStoref            );

  OPENGL_LOAD_GL_FN( glPixelStorei            );

  OPENGL_LOAD_GL_FN( glPixelTransferf         );

  OPENGL_LOAD_GL_FN( glPixelTransferi         );

  OPENGL_LOAD_GL_FN( glPixelZoom              );

  OPENGL_LOAD_GL_FN( glPointSize              );

  OPENGL_LOAD_GL_FN( glPolygonMode            );

  OPENGL_LOAD_GL_FN( glPolygonOffset          );

  OPENGL_LOAD_GL_FN( glPolygonStipple         );

  OPENGL_LOAD_GL_FN( glPopAttrib              );

  OPENGL_LOAD_GL_FN( glPopClientAttrib        );

  OPENGL_LOAD_GL_FN( glPopMatrix              );

  OPENGL_LOAD_GL_FN( glPopName                );

  OPENGL_LOAD_GL_FN( glPrioritizeTextures     );

  OPENGL_LOAD_GL_FN( glPushAttrib             );

  OPENGL_LOAD_GL_FN( glPushClientAttrib       );

  OPENGL_LOAD_GL_FN( glPushMatrix             );

  OPENGL_LOAD_GL_FN( glPushName               );

  OPENGL_LOAD_GL_FN( glRasterPos2d            );

  OPENGL_LOAD_GL_FN( glRasterPos2dv           );

  OPENGL_LOAD_GL_FN( glRasterPos2f            );

  OPENGL_LOAD_GL_FN( glRasterPos2fv           );

  OPENGL_LOAD_GL_FN( glRasterPos2i            );

  OPENGL_LOAD_GL_FN( glRasterPos2iv           );

  OPENGL_LOAD_GL_FN( glRasterPos2s            );

  OPENGL_LOAD_GL_FN( glRasterPos2sv           );

  OPENGL_LOAD_GL_FN( glRasterPos3d            );

  OPENGL_LOAD_GL_FN( glRasterPos3dv           );

  OPENGL_LOAD_GL_FN( glRasterPos3f            );

  OPENGL_LOAD_GL_FN( glRasterPos3fv           );

  OPENGL_LOAD_GL_FN( glRasterPos3i            );

  OPENGL_LOAD_GL_FN( glRasterPos3iv           );

  OPENGL_LOAD_GL_FN( glRasterPos3s            );

  OPENGL_LOAD_GL_FN( glRasterPos3sv           );

  OPENGL_LOAD_GL_FN( glRasterPos4d            );

  OPENGL_LOAD_GL_FN( glRasterPos4dv           );

  OPENGL_LOAD_GL_FN( glRasterPos4f            );

  OPENGL_LOAD_GL_FN( glRasterPos4fv           );

  OPENGL_LOAD_GL_FN( glRasterPos4i            );

  OPENGL_LOAD_GL_FN( glRasterPos4iv           );

  OPENGL_LOAD_GL_FN( glRasterPos4s            );

  OPENGL_LOAD_GL_FN( glRasterPos4sv           );

  OPENGL_LOAD_GL_FN( glReadBuffer             );

  OPENGL_LOAD_GL_FN( glReadPixels             );

  OPENGL_LOAD_GL_FN( glRectd                  );

  OPENGL_LOAD_GL_FN( glRectdv                 );

  OPENGL_LOAD_GL_FN( glRectf                  );

  OPENGL_LOAD_GL_FN( glRectfv                 );

  OPENGL_LOAD_GL_FN( glRecti                  );

  OPENGL_LOAD_GL_FN( glRectiv                 );

  OPENGL_LOAD_GL_FN( glRects                  );

  OPENGL_LOAD_GL_FN( glRectsv                 );

  OPENGL_LOAD_GL_FN( glRenderMode             );

  OPENGL_LOAD_GL_FN( glRotated                );

  OPENGL_LOAD_GL_FN( glRotatef                );

  OPENGL_LOAD_GL_FN( glScaled                 );

  OPENGL_LOAD_GL_FN( glScalef                 );

  OPENGL_LOAD_GL_FN( glScissor                );

  OPENGL_LOAD_GL_FN( glSelectBuffer           );

  OPENGL_LOAD_GL_FN( glShadeModel             );

  OPENGL_LOAD_GL_FN( glStencilFunc            );

  OPENGL_LOAD_GL_FN( glStencilMask            );

  OPENGL_LOAD_GL_FN( glStencilOp              );

  OPENGL_LOAD_GL_FN( glTexCoord1d             );

  OPENGL_LOAD_GL_FN( glTexCoord1dv            );

  OPENGL_LOAD_GL_FN( glTexCoord1f             );

  OPENGL_LOAD_GL_FN( glTexCoord1fv            );

  OPENGL_LOAD_GL_FN( glTexCoord1i             );

  OPENGL_LOAD_GL_FN( glTexCoord1iv            );

  OPENGL_LOAD_GL_FN( glTexCoord1s             );

  OPENGL_LOAD_GL_FN( glTexCoord1sv            );

  OPENGL_LOAD_GL_FN( glTexCoord2d             );

  OPENGL_LOAD_GL_FN( glTexCoord2dv            );

  OPENGL_LOAD_GL_FN( glTexCoord2f             );

  OPENGL_LOAD_GL_FN( glTexCoord2fv            );

  OPENGL_LOAD_GL_FN( glTexCoord2i             );

  OPENGL_LOAD_GL_FN( glTexCoord2iv            );

  OPENGL_LOAD_GL_FN( glTexCoord2s             );

  OPENGL_LOAD_GL_FN( glTexCoord2sv            );

  OPENGL_LOAD_GL_FN( glTexCoord3d             );

  OPENGL_LOAD_GL_FN( glTexCoord3dv            );

  OPENGL_LOAD_GL_FN( glTexCoord3f             );

  OPENGL_LOAD_GL_FN( glTexCoord3fv            );

  OPENGL_LOAD_GL_FN( glTexCoord3i             );

  OPENGL_LOAD_GL_FN( glTexCoord3iv            );

  OPENGL_LOAD_GL_FN( glTexCoord3s             );

  OPENGL_LOAD_GL_FN( glTexCoord3sv            );

  OPENGL_LOAD_GL_FN( glTexCoord4d             );

  OPENGL_LOAD_GL_FN( glTexCoord4dv            );

  OPENGL_LOAD_GL_FN( glTexCoord4f             );

  OPENGL_LOAD_GL_FN( glTexCoord4fv            );

  OPENGL_LOAD_GL_FN( glTexCoord4i             );

  OPENGL_LOAD_GL_FN( glTexCoord4iv            );

  OPENGL_LOAD_GL_FN( glTexCoord4s             );

  OPENGL_LOAD_GL_FN( glTexCoord4sv            );

  OPENGL_LOAD_GL_FN( glTexCoordPointer        );

  OPENGL_LOAD_GL_FN( glTexEnvf                );

  OPENGL_LOAD_GL_FN( glTexEnvfv               );

  OPENGL_LOAD_GL_FN( glTexEnvi                );

  OPENGL_LOAD_GL_FN( glTexEnviv               );

  OPENGL_LOAD_GL_FN( glTexGend                );

  OPENGL_LOAD_GL_FN( glTexGendv               );

  OPENGL_LOAD_GL_FN( glTexGenf                );

  OPENGL_LOAD_GL_FN( glTexGenfv               );

  OPENGL_LOAD_GL_FN( glTexGeni                );

  OPENGL_LOAD_GL_FN( glTexGeniv               );

  OPENGL_LOAD_GL_FN( glTexImage1D             );

  OPENGL_LOAD_GL_FN( glTexImage2D             );

  OPENGL_LOAD_GL_FN( glTexParameterf          );

  OPENGL_LOAD_GL_FN( glTexParameterfv         );

  OPENGL_LOAD_GL_FN( glTexParameteri          );

  OPENGL_LOAD_GL_FN( glTexParameteriv         );

  OPENGL_LOAD_GL_FN( glTexSubImage1D          );

  OPENGL_LOAD_GL_FN( glTexSubImage2D          );

  OPENGL_LOAD_GL_FN( glTranslated             );

  OPENGL_LOAD_GL_FN( glTranslatef             );

  OPENGL_LOAD_GL_FN( glVertex2d               );

  OPENGL_LOAD_GL_FN( glVertex2dv              );

  OPENGL_LOAD_GL_FN( glVertex2f               );

  OPENGL_LOAD_GL_FN( glVertex2fv              );

  OPENGL_LOAD_GL_FN( glVertex2i               );

  OPENGL_LOAD_GL_FN( glVertex2iv              );

  OPENGL_LOAD_GL_FN( glVertex2s               );

  OPENGL_LOAD_GL_FN( glVertex2sv              );

  OPENGL_LOAD_GL_FN( glVertex3d               );

  OPENGL_LOAD_GL_FN( glVertex3dv              );

  OPENGL_LOAD_GL_FN( glVertex3f               );

  OPENGL_LOAD_GL_FN( glVertex3fv              );

  OPENGL_LOAD_GL_FN( glVertex3i               );

  OPENGL_LOAD_GL_FN( glVertex3iv              );

  OPENGL_LOAD_GL_FN( glVertex3s               );

  OPENGL_LOAD_GL_FN( glVertex3sv              );

  OPENGL_LOAD_GL_FN( glVertex4d               );

  OPENGL_LOAD_GL_FN( glVertex4dv              );

  OPENGL_LOAD_GL_FN( glVertex4f               );

  OPENGL_LOAD_GL_FN( glVertex4fv              );

  OPENGL_LOAD_GL_FN( glVertex4i               );

  OPENGL_LOAD_GL_FN( glVertex4iv              );

  OPENGL_LOAD_GL_FN( glVertex4s               );

  OPENGL_LOAD_GL_FN( glVertex4sv              );

  OPENGL_LOAD_GL_FN( glVertexPointer          );

  OPENGL_LOAD_GL_FN( glViewport               );

#endif



  if (openglGLUInst) {

    OPENGL_LOAD_GLU_FN( gluErrorString );

    OPENGL_LOAD_GLU_FN( gluErrorUnicodeStringEXT );

    OPENGL_LOAD_GLU_FN( gluGetString );

    OPENGL_LOAD_GLU_FN( gluOrtho2D );

    OPENGL_LOAD_GLU_FN( gluPerspective );

    OPENGL_LOAD_GLU_FN( gluPickMatrix );

    OPENGL_LOAD_GLU_FN( gluLookAt );

    OPENGL_LOAD_GLU_FN( gluProject );

    OPENGL_LOAD_GLU_FN( gluUnProject );

    OPENGL_LOAD_GLU_FN( gluScaleImage );

    OPENGL_LOAD_GLU_FN( gluBuild1DMipmaps );

    OPENGL_LOAD_GLU_FN( gluBuild2DMipmaps );

    OPENGL_LOAD_GLU_FN( gluNewQuadric );

    OPENGL_LOAD_GLU_FN( gluDeleteQuadric );

    OPENGL_LOAD_GLU_FN( gluQuadricNormals );

    OPENGL_LOAD_GLU_FN( gluQuadricTexture );

    OPENGL_LOAD_GLU_FN( gluQuadricOrientation );

    OPENGL_LOAD_GLU_FN( gluQuadricDrawStyle );

    OPENGL_LOAD_GLU_FN( gluCylinder );

    OPENGL_LOAD_GLU_FN( gluDisk );

    OPENGL_LOAD_GLU_FN( gluPartialDisk );

    OPENGL_LOAD_GLU_FN( gluSphere );

    OPENGL_LOAD_GLU_FN( gluQuadricCallback );

    OPENGL_LOAD_GLU_FN( gluNewTess  );   

    OPENGL_LOAD_GLU_FN( gluDeleteTess );

    OPENGL_LOAD_GLU_FN( gluTessBeginPolygon );

    OPENGL_LOAD_GLU_FN( gluTessBeginContour );

    OPENGL_LOAD_GLU_FN( gluTessVertex );

    OPENGL_LOAD_GLU_FN( gluTessEndContour );

    OPENGL_LOAD_GLU_FN( gluTessEndPolygon );

    OPENGL_LOAD_GLU_FN( gluTessProperty );  

    OPENGL_LOAD_GLU_FN( gluTessNormal );

    OPENGL_LOAD_GLU_FN( gluTessCallback );  

    OPENGL_LOAD_GLU_FN( gluGetTessProperty );

    OPENGL_LOAD_GLU_FN( gluNewNurbsRenderer );

    OPENGL_LOAD_GLU_FN( gluDeleteNurbsRenderer );

    OPENGL_LOAD_GLU_FN( gluBeginSurface );

    OPENGL_LOAD_GLU_FN( gluBeginCurve );

    OPENGL_LOAD_GLU_FN( gluEndCurve );

    OPENGL_LOAD_GLU_FN( gluEndSurface );

    OPENGL_LOAD_GLU_FN( gluBeginTrim );

    OPENGL_LOAD_GLU_FN( gluEndTrim );

    OPENGL_LOAD_GLU_FN( gluPwlCurve );

    OPENGL_LOAD_GLU_FN( gluNurbsCurve );

    OPENGL_LOAD_GLU_FN( gluNurbsSurface );  

    OPENGL_LOAD_GLU_FN( gluLoadSamplingMatrices );

    OPENGL_LOAD_GLU_FN( gluNurbsProperty );

    OPENGL_LOAD_GLU_FN( gluGetNurbsProperty );

    OPENGL_LOAD_GLU_FN( gluNurbsCallback );

  }



  /* 

   * Anything other than an ICD/MCD full gl driver bypasses the GDI entirely

   */

  if (strcmp(lib, "opengl32.dll")) openglBypassGDI = TRUE;



  return 0; 

}



//////////////////////////////////////////////////////////////////////////



void OpenGLLoadExtensions(void) {

  /* extension functions - check that the extension is supported before using these */

  //////////////// For GL_ARB_multitexture extension. //////////////////////

  OPENGL_LOAD_EXT_FN( glActiveTextureARB        );

  OPENGL_LOAD_EXT_FN( glClientActiveTextureARB  );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord1dARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord1dvARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord1fARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord1fvARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord1iARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord1ivARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord1sARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord1svARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord2dARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord2dvARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord2fARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord2fvARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord2iARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord2ivARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord2sARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord2svARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord3dARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord3dvARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord3fARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord3fvARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord3iARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord3ivARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord3sARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord3svARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord4dARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord4dvARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord4fARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord4fvARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord4iARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord4ivARB     );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord4sARB      );

  OPENGL_LOAD_EXT_FN( glMultiTexCoord4svARB     );



  ///////////////// For GL_EXT_compiled_vertex_array extension. ////////////

  OPENGL_LOAD_EXT_FN( glLockArraysEXT           );

  OPENGL_LOAD_EXT_FN( glUnlockArraysEXT         );



  ///////////////// For WGL_EXT_swap_control extension. ////////////

  OPENGL_LOAD_EXT_FN( wglSwapIntervalEXT        );

  OPENGL_LOAD_EXT_FN( wglGetSwapIntervalEXT     );



  ///////////////// For GL_EXT_point_parameters extension. ////////////

  OPENGL_LOAD_EXT_FN( glPointParameterfEXT      );

  OPENGL_LOAD_EXT_FN( glPointParameterfvEXT     );

}



//////////////////////////////////////////////////////////////////////////



void OpenGLUnInit(void) {

  if (!openglInit) return;

  openglInit-=1;

  if (openglInit) return;



  if (openglGLUInst) FreeLibrary(openglInst); 

  openglInst = NULL; 

  if (openglGLUInst) FreeLibrary(openglGLUInst); 

  openglGLUInst = NULL; 

} 





//////////////////////////////////////////////////////////////////////////



/* Determine if an extension is supported, must be called after a valid

 * context has been created */

GLboolean OpenGLCheckExtension(char *extName) {

  char * extensions;

  char * substring;

  char   lastChar;



  if (!fnglGetString) return GL_FALSE;

  extensions = (char *)fnglGetString(GL_EXTENSIONS);

  /* This is bad, because we can find extensions that are substrings of other extensions */

  substring = strstr(extensions, extName);

  if (!substring) return GL_FALSE;

  lastChar = substring[ strlen(extName) ];

  if (lastChar==' ' || lastChar==0) return GL_TRUE;

  /* anything else means its a substring of a longer extension */

  return GL_FALSE;

}

  

/******************************************************************************/

/* replaced WGL* functions */



/*

 * These cant? be fixed like the above functions becuase they are declared in gdi.h

 * and for some reason the compiler will insist on declaring them as EXPORTED

 * dll functions... weird

 */



BOOL OpenGLCopyContext(HGLRC hglrcSrc, HGLRC hlglrcDst, UINT mask) {

  BOOL (APIENTRY *fnwglCopyContext)(HGLRC hglrcSrc, HGLRC hlglrcDst, UINT mask);



  if (!openglInst) return FALSE;

  fnwglCopyContext = (void*)GetProcAddress(openglInst, "wglCopyContext");

  if (!fnwglCopyContext) return FALSE;

  return (*fnwglCopyContext)(hglrcSrc, hlglrcDst, mask);

}





HGLRC OpenGLCreateContext(HDC hdc) {

  HGLRC (APIENTRY *fnwglCreateContext)(HDC hdc);



  if (!openglInst) return NULL;

  fnwglCreateContext = (void*)GetProcAddress(openglInst, "wglCreateContext");

  if (!fnwglCreateContext) return NULL;

  return (*fnwglCreateContext)(hdc);

}





//HGLRC     OpenGLCreateLayerContext(HDC, int);



BOOL OpenGLDeleteContext(HGLRC hglrc) {

  BOOL (APIENTRY *fnwglDeleteContext)(HGLRC hglrc);



  if (!openglInst) return FALSE;

  fnwglDeleteContext = (void*)GetProcAddress(openglInst, "wglDeleteContext");

  if (!fnwglDeleteContext) return FALSE;

  return (*fnwglDeleteContext)(hglrc);

}





//HGLRC     OpenGLGetCurrentContext(VOID);

//HDC       OpenGLGetCurrentDC(VOID);



PROC OpenGLGetProcAddress(LPCSTR lpcstr) {

  PROC (APIENTRY *fnwglGetProcAddress)(LPCSTR lpcstr);

  PROC  fn=NULL;



  if(!openglInst) return FALSE;

  fnwglGetProcAddress = (void *)GetProcAddress(openglInst, "wglGetProcAddress");

  if(!fnwglGetProcAddress) return NULL;

  fn = (*fnwglGetProcAddress)(lpcstr);

  if (!fn) fn = (void*)GetProcAddress(openglInst, lpcstr);

  return fn;

}





BOOL OpenGLMakeCurrent(HDC hdc, HGLRC hglrc) {

  BOOL (APIENTRY *fnwglMakeCurrent)(HDC hdc, HGLRC hglrc);

  BOOL retValue;



  if (!openglInst) return FALSE;

  fnwglMakeCurrent = (void*)GetProcAddress(openglInst, "wglMakeCurrent");

  if (!fnwglMakeCurrent) return FALSE;

  retValue = (*fnwglMakeCurrent)(hdc, hglrc);

#if (AUTOLOAD_EXTENSIONS)

  if (retValue) {

    static HGLRC lastHglrc = NULL;

    if (hglrc != lastHglrc) {

      lastHglrc = hglrc;

      OpenGLLoadExtensions();

    }

  }

#endif

  return retValue;

}



//BOOL      OpenGLShareLists(HGLRC, HGLRC);



BOOL OpenGLUseFontBitmapsW(HDC hdc, DWORD first, DWORD count, DWORD listBase) {

  BOOL (APIENTRY *fnwglUseFontBitmaps)(HDC hdc, DWORD first, DWORD count, DWORD listBase);



  if (!openglInst) return FALSE;

  fnwglUseFontBitmaps = (void*)GetProcAddress(openglInst, "wglUseFontBitmapsW");

  if (!fnwglUseFontBitmaps) return FALSE;

  return (*fnwglUseFontBitmaps)(hdc, first, count, listBase);

}





BOOL OpenGLUseFontBitmapsA(HDC hdc, DWORD first, DWORD count, DWORD listBase) {

  BOOL (APIENTRY *fnwglUseFontBitmaps)(HDC hdc, DWORD first, DWORD count, DWORD listBase);



  if (!openglInst) return FALSE;

  fnwglUseFontBitmaps = (void*)GetProcAddress(openglInst, "wglUseFontBitmapsA");

  if (!fnwglUseFontBitmaps) return FALSE;

  return (*fnwglUseFontBitmaps)(hdc, first, count, listBase);

}





BOOL OpenGLSwapLayerBuffers(HDC hdc, UINT fuPlanes) {

  BOOL (APIENTRY *fnwglSwapLayerBuffers)(HDC hdc, UINT fuPlanes);



  if (!openglInst) return FALSE;

  fnwglSwapLayerBuffers = (void*)GetProcAddress(openglInst, "wglSwapLayerBuffers");

  if (!fnwglSwapLayerBuffers) return FALSE;

  return (*fnwglSwapLayerBuffers)(hdc, fuPlanes);

}





/******************************************************************************/

/* replaced GDI functions */

/* 

 * When using MiniGL drivers you can't call the GDI functions (ie SwapBuffers etc)

 * because they will redirect the call to the system ogl dll lib, not the minigl dll

 * that you are using.

 */



BOOL OpenGLSwapBuffers(HDC hdc) {

  BOOL (APIENTRY *fnSwapBuffers)(HDC hdc);



  if (!openglBypassGDI) {

    // Try GDI version first

    int retValue;

    retValue = SwapBuffers(hdc);

    if (retValue) return retValue;

  }

  if (!openglInst) return FALSE;

  fnSwapBuffers = (void*)GetProcAddress(openglInst, "wglSwapBuffers");

  if (!fnSwapBuffers) return FALSE;

  return (*fnSwapBuffers)(hdc);

}



int OpenGLChoosePixelFormat(HDC hdc, CONST PIXELFORMATDESCRIPTOR * ppfd) {

  int (APIENTRY *fnChoosePixelFormat)(HDC hdc, CONST PIXELFORMATDESCRIPTOR * ppfd);



  if (!openglBypassGDI) {

    // Try GDI version first

    int retValue;

    retValue = ChoosePixelFormat(hdc, ppfd);

    if (retValue) return retValue;

  }

  // Bypass GDI entirely

  if (!openglInst) return FALSE;

  fnChoosePixelFormat = (void*)GetProcAddress(openglInst, "wglChoosePixelFormat");

  if (!fnChoosePixelFormat) return FALSE;

  openglBypassGDI = TRUE;

  return (*fnChoosePixelFormat)(hdc, ppfd); 

}



int OpenGLDescribePixelFormat(HDC hdc, int iPixelFormat, UINT nBytes, LPPIXELFORMATDESCRIPTOR ppfd) {

  int (APIENTRY *fnDescribePixelFormat)(HDC hdc, int iPixelFormat, UINT nBytes, LPPIXELFORMATDESCRIPTOR ppfd);



  if (!openglBypassGDI) {

    // Try GDI version first

    int retValue;

    retValue = DescribePixelFormat(hdc, iPixelFormat, nBytes, ppfd);

    if (retValue) return retValue;

  }

  // Bypass GDI entirely

  if (!openglInst) return FALSE;

  fnDescribePixelFormat = (void*)GetProcAddress(openglInst, "wglDescribePixelFormat");

  if (!fnDescribePixelFormat) return FALSE;

  openglBypassGDI = TRUE;

  return (*fnDescribePixelFormat)(hdc, iPixelFormat, nBytes, ppfd);

}



BOOL OpenGLSetPixelFormat(HDC hdc, int iPixelFormat, CONST PIXELFORMATDESCRIPTOR * ppfd) {

  BOOL (APIENTRY *fnSetPixelFormat)(HDC hdc, int iPixelFormat, CONST PIXELFORMATDESCRIPTOR * ppfd);

  int retValue = 0;



  if (!openglBypassGDI) {

    // Try GDI version first

    retValue = SetPixelFormat(hdc, iPixelFormat, ppfd);

  }

  // Bypass GDI entirely

  if (!retValue) {

    if (!openglInst) return FALSE;

    fnSetPixelFormat = (void*)GetProcAddress(openglInst, "wglSetPixelFormat");

    if (!fnSetPixelFormat) return FALSE;

    openglBypassGDI = TRUE;

    retValue = (*fnSetPixelFormat)(hdc, iPixelFormat, ppfd);

  }

  return retValue;

}



