/*
 * This code should work fine with either a full opengl DLL or
 * the opengl32.dll from a miniport.
 *
 * This program demonstrates how to convert a program from staticly
 * binding to the opengl driver to using my dynamic gl loading code
 *
 * Its a edited excerpt from my own "plasma" game engine, currently
 * only used in my Azteroidz3D game demo. (If you're interested you
 * can get it from my web page, the client/server coop multiplay
 * version should be posted soon)
 *
 * The output of this program should be two spinning cubes.
 *
 * You can email me: cryogen@infoserve.net
 * Visit my web page: http://home.bc.rogers.wave.ca/borealis/ryan.html
 * 
 */
/*
 * PS: Thanx to Doug @ Nvidia Developer relations for the pragma idea, 
 *    it makes things much more idiot-proof 
 * 
 * Thanx also to Parrish Heywood, who pointed out that I was asleep when I
 * came up with the old glFrustrum calculation. Ergh. But its much better now....
 */

/* 
 * You must also link to the following libs to get this to work
 * winmm.lib
 */
#pragma comment(lib, "winmm.lib")

/* turn off warning: 4018 signed/unsigned mismatch */
#pragma warning( disable : 4018)


#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

/* 
 * It also requires my dynamic opengl dll loading code
 * opengl.c opengl.h
 */
#include "opengl.h"

#define LOG_BOOT        0
#define LOG_ERROR       1

#define WIDTH           640
#define HEIGHT          480
#define BITDEPTH        16
#define FULLSCREEN      1
#define PRIORITY        1 

/* 
 * not much of a test with just 2 cubes, but you get the idea.
 * Handy for testing the improvements of engine mods once you get going.
 */
#define FRAMETEST       0     /* wanna frame test yer card? enable this */
#define TESTTIME        30.0f /* test for 30 secs */

/* in degrees per second */
#define YAW_RATE        45.0f
#define PITCH_RATE      45.0f
#define ROLL_RATE       45.0f

#define PI              3.141592653589793

#define BYTE            unsigned char

#define Float(x)        ((float)(x))

CHAR  szAppName[]="GLDemo";
HDC   hDC;
HWND  hWndMain;
HGLRC hRC;
LONG  fps;
LONG  priority=PRIORITY;

/* Should be in a struct of course, but just a demo */
LONG texID = 1; /* next texture would be given ID 2, then 3 etc */
LONG texWidth = 4;
LONG texHeight = 4;
BYTE texImage[] = { /* rgba 8 bits per pixel */
  64, 255, 0,   255, /* 0,0 */
  64, 255, 0,   255, /* 1,0 */
  64, 64,  255, 255, /* 2,0 */
  64, 255, 0,   255, /* 3,0 */
  64, 255, 0,   255, /* 0,1 */
  64, 64,  255, 255, /* 1,1 */
  64, 255, 0,   255, /* 2,1 */
  64, 255, 0,   255, /* 3,1 */
  64, 64,  255, 255, /* 0,2 */
  64, 255, 0,   255, /* 1,2 */
  64, 64,  255, 255, /* 2,2 */
  64, 64,  255, 255, /* 3,2 */
  64, 255, 0,   255, /* 0,3 */
  64, 64,  255, 255, /* 1,3 */
  64, 255, 0,   255, /* 2,3 */
  64, 255, 0,   255  /* 3,3 */
};

/* A note on opengl coordinate axes (which are "left-handed")*/
/*
 *     Y
 *     ^      *MINUS* Z
 *     |    /
 *     |  /
 *     |/
 *     +------> X
 *
 * So positive Z is coming out of the screen, something which I find
 * counter-intuitive.
 */

/* Typically these would be structs but lets keep this simple */
float cameraYaw    = 0.0f;
float cameraPitch  = 0.0f;
float cameraRoll   = 0.0f;
float cameraX      = 0.0f;
float cameraY      = 0.0f;
float cameraZ      = 8.0f;
float fovX         = 90.0f;

float rgbCubeYaw   = 0.0f;
float rgbCubePitch = 0.0f;
float rgbCubeRoll  = 0.0f;
float rgbCubeX     =-2.0f;
float rgbCubeY     = 0.0f;
float rgbCubeZ     = 0.0f;
float rgbCubeScale = 1.0f;

float texCubeYaw   = 0.0f;
float texCubePitch = 0.0f;
float texCubeRoll  = 0.0f;
float texCubeX     = 2.0f;
float texCubeY     = 0.0f;
float texCubeZ     = 0.0f;
float texCubeScale = 1.0f;

/* I log everything to file normally, but no need for the purposes of this demo */
void __cdecl Log(INT32 logType, const BYTE *logStr, ...) {
  va_list     varList;
  BYTE        buf[1024];

  va_start( varList, logStr );     /* Initialize variable arguments. */
  if (logType==LOG_BOOT) {
    OutputDebugString("BT!:");
  } else {
    OutputDebugString("ERR:");
  }
  vsprintf(buf, logStr, varList);
  OutputDebugString(buf);
}

void LogPrintf(INT32 logType, BYTE *logStr) {
  OutputDebugString(logStr);
}


/* 
 * As far as I'm concerned this should be a standard function call, why I have to 
 * waste my time wrapping functions like this I cant tell ya
 */
void ShowSystemError(char *error) {
  LPTSTR lpMsgBuf;

  FormatMessage( 
      FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
      NULL,
      GetLastError(),
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
      (LPTSTR) &lpMsgBuf,
      0,
      NULL 
  );

  // Display the string.
  Log(LOG_ERROR, lpMsgBuf);
  LogPrintf(LOG_ERROR, "\n");

  if (error)
    MessageBox( NULL, lpMsgBuf, error, MB_OK|MB_ICONINFORMATION );
  else
    MessageBox( NULL, lpMsgBuf, "GetLastError", MB_OK|MB_ICONINFORMATION );

  // Free the buffer.
  LocalFree( lpMsgBuf );
}

void ListDisplaySettings(void) {
  DEVMODE devMode;
  LONG    i;
  LONG    modeExist;
  BYTE    buf[1024];

  /* enumerate all modes to log */
  modeExist = EnumDisplaySettings(NULL, 0, &devMode);
  sprintf(buf, "Available screen modes for %s:\n", devMode.dmDeviceName);
  Log(LOG_BOOT, buf);
  for (i=0; modeExist;i++) {
    sprintf(buf, "%ldx%ldx%ld-bit color (%ld hertz)\n", devMode.dmPelsWidth, devMode.dmPelsHeight, devMode.dmBitsPerPel, devMode.dmDisplayFrequency);
    Log(LOG_BOOT, buf);

    modeExist = EnumDisplaySettings(NULL, i, &devMode);
  }
}

/* 
 * Need to change the pixel format / buffer of the
 * window to what we need for the opengl fullscreen
 */
BOOL ChangeResolution(LONG  w, LONG  h, LONG  bitdepth) {
  DEVMODE devMode;
  LONG    modeExist;
  LONG    modeSwitch;
  LONG    closeMode = 0;
  LONG    i;
  BYTE    buf[256];

  for (i=0; ;i++) {
    modeExist = EnumDisplaySettings(NULL, i, &devMode);

    /* Check if we ran out of modes */
    if (!modeExist) {
      ListDisplaySettings();

      /* Pony up an error message */
      sprintf(buf, "No matching mode to %ldx%ldx%ld-bit color available\n", w, h);
      Log(LOG_BOOT, buf);
      break;
    }

    /* Check for a matching mode */
    if ( (devMode.dmBitsPerPel == bitdepth) 
       &&(devMode.dmPelsWidth == w) 
       &&(devMode.dmPelsHeight == h) )
    {
      sprintf(buf, "Trying %ldx%ldx%ld-bit color (%ld hertz)\n", devMode.dmPelsWidth, devMode.dmPelsHeight, devMode.dmBitsPerPel, devMode.dmDisplayFrequency);
      Log(LOG_BOOT, buf);
      modeSwitch = ChangeDisplaySettings(&devMode, CDS_FULLSCREEN); 
      if (modeSwitch==DISP_CHANGE_SUCCESSFUL) return TRUE;
      Log(LOG_BOOT, "Nope, didnt work\n");
      if (!closeMode) closeMode = i;
    }
  }

  //Might be running in Windows95, let's try without the hertz change 
  EnumDisplaySettings(NULL, closeMode, &devMode);
  devMode.dmBitsPerPel = bitdepth; 
  devMode.dmPelsWidth  = w; 
  devMode.dmPelsHeight = h; 
  devMode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT; 
  sprintf(buf, "Trying %ldx%ldx%ld-bit color (Win95 Hack)\n", devMode.dmPelsWidth, devMode.dmPelsHeight, devMode.dmBitsPerPel, devMode.dmDisplayFrequency);
  Log(LOG_BOOT, buf);
  modeSwitch = ChangeDisplaySettings(&devMode, CDS_FULLSCREEN); 
  if(modeSwitch == DISP_CHANGE_SUCCESSFUL) return TRUE;
  Log(LOG_BOOT, "Nope, didnt work\n");

  /* try with separate changes to bitdepth and resolution */
  devMode.dmFields = DM_BITSPERPEL; 
  sprintf(buf, "Trying %ldx%ldx%ld-bit color (Separate BPP switch)\n", devMode.dmPelsWidth, devMode.dmPelsHeight, devMode.dmBitsPerPel, devMode.dmDisplayFrequency);
  Log(LOG_BOOT, buf);
  modeSwitch = ChangeDisplaySettings(&devMode, CDS_FULLSCREEN);
  if(modeSwitch == DISP_CHANGE_SUCCESSFUL) {
    devMode.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT; 
    sprintf(buf, "Trying %ldx%ldx%ld-bit color (Separate resolution switch)\n", devMode.dmPelsWidth, devMode.dmPelsHeight, devMode.dmBitsPerPel, devMode.dmDisplayFrequency);
    Log(LOG_BOOT, buf);
    modeSwitch = ChangeDisplaySettings(&devMode, CDS_FULLSCREEN); 
    if(modeSwitch == DISP_CHANGE_SUCCESSFUL) return TRUE;
    ChangeDisplaySettings(NULL, 0);
  }
  Log(LOG_BOOT, "Nope, didnt work\n");

  /* Well nothing works now does it */
  if(modeSwitch == DISP_CHANGE_RESTART) { 
    sprintf(buf, "You must change your video mode to %ldx%ldx%ld-bit color and then restart\n", w, h, bitdepth);
    Log(LOG_BOOT, buf);
  } else if(modeSwitch == DISP_CHANGE_BADMODE) {
    sprintf(buf, "The video mode %ldx%ldx%ld-bit color is not supported\n", w, h, bitdepth);
    Log(LOG_BOOT, buf);
  } else if(modeSwitch == DISP_CHANGE_FAILED) {
    sprintf(buf, "Hardware failed to change to %ldx%ldx%ld-bit color\n", w, h, bitdepth);
    Log(LOG_BOOT, buf);
  } else {
    sprintf(buf, "Failed to change to %ldx%ldx%ld-bit color - unknown reason\n", w, h, bitdepth);
    Log(LOG_BOOT, buf);
  }
  return FALSE; 
}

void SetPriority(LONG priority) {
  HANDLE    hProc;
  hProc = GetCurrentProcess();

  if (priority==0) {
    SetPriorityClass(hProc, IDLE_PRIORITY_CLASS);
  } else if (priority==1) {
    SetPriorityClass(hProc, NORMAL_PRIORITY_CLASS);
  } else if (priority==3) {
    SetPriorityClass(hProc, REALTIME_PRIORITY_CLASS);
  } else {
    SetPriorityClass(hProc, HIGH_PRIORITY_CLASS);
  }
}

BOOL SetupPixelFormat(HDC hdc, int colorDepth, int zDepth) {
  PIXELFORMATDESCRIPTOR pfd;
  int                   pixelformat;
  BOOL                  success;

  memset(&pfd, 0, sizeof(pfd));
  pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
  pfd.nVersion = 1;
  pfd.dwFlags = PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW | PFD_DOUBLEBUFFER;
  pfd.dwLayerMask = PFD_MAIN_PLANE;
  pfd.iPixelType = PFD_TYPE_RGBA;
  pfd.cColorBits = colorDepth;
  pfd.cDepthBits = zDepth;
  pfd.cAccumBits = 0;
  pfd.cStencilBits = 0;

  pixelformat = OpenGLChoosePixelFormat(hdc, &pfd);
  if (pixelformat == 0) {
    Log(LOG_BOOT, "ChoosePixelFormat failed\n");
    ShowSystemError("ChoosePixelFormat failed");
    return FALSE;
  }

#ifdef CURIOUS_ABOUT_FORMAT
  /* what did we get I wonder */
  memset(&pfd, 0, sizeof(pfd));
  pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
  success = OpenGLDescribePixelFormat(hdc, pixelformat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
  if (! success ) {
    Log(LOG_BOOT, "DescribePixelFormat failed\n");
    ShowSystemError("DescribePixelFormat failed");
    return FALSE;
  }
#endif

  /* set the format */
  success = OpenGLSetPixelFormat(hdc, pixelformat, &pfd);
  if (! success ) {
    Log(LOG_BOOT, "SetPixelFormat failed\n");
    MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
    ShowSystemError("SetPixelFormat failed");
    return FALSE;
  }

  return TRUE;
}


void ResizeViewport( LONG  width, LONG  height ) {
  double  znear = 0.7; /* Put the screen 1 unit in front of the "eye" */
  double  zfar  = 250.0; /* Note that the ratio far/near is to be minimized */
  double  aspect;

  /* Set a viewport to the entire screen, its all one big 3D window now */
  glViewport(0, 0, width, height);

  /* 
   * Calculate the screen (ie a neutral) aspect ratio 
   * ie squares are squares 
   */
  aspect = (GLdouble)width/(GLdouble)height;

  /* Set the Projection matrix */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

   /* gluPerspective sets field of view for Y, but since I want a known
   * (ie 90 degree) field of view in the x direction calc our own using
   * raw frustum parameters instead.
   */
#define USE_FRUSTUM 1
#if (USE_FRUSTUM)
  {
    double  left;
    double  right;
    double  bottom;
    double  top;

    right = znear * tan( fovX/2.0 * PI/180.0f );
    top = right / aspect;
    bottom = -top;
    left = -right;
    glFrustum(left, right, bottom, top, znear, zfar);
  }
#else
  gluPerspective(fovX/aspect, aspect, znear, zfar); 
#endif
  glMatrixMode(GL_MODELVIEW);
}    

void InitGL(LONG  width, LONG  height) {
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
  glClearDepth(1.0);
  glDepthFunc(GL_LEQUAL);  /* Superimpose smaller Z values over larger ones */
  glShadeModel(GL_SMOOTH); /* Smooth shading */
  glFrontFace(GL_CCW);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glAlphaFunc(GL_GEQUAL, 0.07f);
  glDisable(GL_ALPHA_TEST);
  glDisable(GL_BLEND);
  glDisable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST); /* Z buffering is go... */
  
  //glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST|GL_FASTEST);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

  ResizeViewport(width, height);
  glMatrixMode(GL_MODELVIEW);


  /* 
   * Note about 3dfx hardware: What texture formats are supported?
   * I think texture size must be a power of 2, good practice anyways
   * 3dfx is also limited to 256x256 resolution textures.
   */
  glBindTexture(GL_TEXTURE_2D, texID);
  
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,     (float)GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,     (float)GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, (float)GL_LINEAR); /* GL_LINEAR would be bilinear filtering */
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, (float)GL_LINEAR); /* GL_LINEAR would be bilinear filtering */
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE,        (float)GL_DECAL);

  glTexImage2D(GL_TEXTURE_2D, 0, 4, texWidth, texHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, texImage );
}

/* Note that order is correct for face-culling */
void DrawTextureCube(GLint texNum) {
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity(); /* Set Model-View to the identity matrix */
  glTranslatef(texCubeX, texCubeY, texCubeZ);
  glScalef(texCubeScale, texCubeScale, texCubeScale);
  glRotatef(texCubePitch, 1.0f, 0.0f, 0.0f); /* rotate about the x unit vector */
  glRotatef(texCubeYaw,   0.0f, 1.0f, 0.0f); /* rotate about the y unit vector */
  glRotatef(texCubeRoll,  0.0f, 0.0f, 1.0f); /* rotate about the z unit vector */

  /* Turn on textures */
  glEnable(GL_TEXTURE_2D);

  /* Draw a textured side of the cube */
  glBindTexture(GL_TEXTURE_2D, texNum);
  glBegin(GL_QUADS);
    glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);
    glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);
    glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);
    glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);

    glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);
    glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);
    glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);
    glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);

    glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f,  1.0f, -1.0f);
    glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);
    glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);
    glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f,  1.0f, -1.0f);

    glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);
    glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f, -1.0f,  1.0f);
    glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f, -1.0f,  1.0f);
    glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);

    glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);
    glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);
    glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);
    glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);

    glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);
    glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);
    glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);
    glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);
  glEnd();
  glDisable(GL_TEXTURE_2D);
}

/*
 * Simple cube with blended colored faces
 */
void DrawRGBCube() {
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity(); /* Set Model-View to the identity matrix */
  glTranslatef(rgbCubeX, rgbCubeY, rgbCubeZ);
  glScalef(rgbCubeScale, rgbCubeScale, rgbCubeScale);
  glRotatef(rgbCubePitch, 1.0f, 0.0f, 0.0f); /* rotate about the x unit vector */
  glRotatef(rgbCubeYaw,   0.0f, 1.0f, 0.0f); /* rotate about the y unit vector */
  glRotatef(rgbCubeRoll,  0.0f, 0.0f, 1.0f); /* rotate about the z unit vector */

  glBegin(GL_QUADS);
    glColor3f(0.0f, 0.0f, 0.0f); /* black corner */
    glVertex3f(-1.0f, -1.0f, -1.0f);
    glColor3f(0.0f, 1.0f, 0.0f); /* green */
    glVertex3f(-1.0f,  1.0f, -1.0f);
    glColor3f(1.0f, 1.0f, 0.0f); /* yellow */
    glVertex3f( 1.0f,  1.0f, -1.0f);
    glColor3f(1.0f, 0.0f, 0.0f); /* red */
    glVertex3f( 1.0f, -1.0f, -1.0f);

    glColor3f(1.0f, 0.0f, 1.0f);  /* purple corner */
    glVertex3f( 1.0f, -1.0f,  1.0f);
    glColor3f(1.0f, 1.0f, 1.0f);  /* white */
    glVertex3f( 1.0f,  1.0f,  1.0f);
    glColor3f(0.0f, 1.0f, 1.0f);  /* cyan */
    glVertex3f(-1.0f,  1.0f,  1.0f);
    glColor3f(0.0f, 0.0f, 1.0f);  /* blue */
    glVertex3f(-1.0f, -1.0f,  1.0f);

    glColor3f(0.0f, 1.0f, 0.0f);  /* green corner */
    glVertex3f(-1.0f,  1.0f, -1.0f);
    glColor3f(0.0f, 1.0f, 1.0f); /* cyan */
    glVertex3f(-1.0f,  1.0f,  1.0f);
    glColor3f(1.0f, 1.0f, 1.0f);  /* white */
    glVertex3f( 1.0f,  1.0f,  1.0f);
    glColor3f(1.0f, 1.0f, 0.0f); /* yellow */
    glVertex3f( 1.0f,  1.0f, -1.0f);

    glColor3f(1.0f, 0.0f, 0.0f); /* red corner */
    glVertex3f( 1.0f, -1.0f, -1.0f);
    glColor3f(1.0f, 0.0f, 1.0f); /* purple */
    glVertex3f( 1.0f, -1.0f,  1.0f);
    glColor3f(0.0f, 0.0f, 1.0f); /* blue */
    glVertex3f(-1.0f, -1.0f,  1.0f);
    glColor3f(0.0f, 0.0f, 0.0f); /* black */
    glVertex3f(-1.0f, -1.0f, -1.0f);

    glColor3f(0.0f, 0.0f, 1.0f); /* blue */
    glVertex3f(-1.0f, -1.0f,  1.0f);
    glColor3f(0.0f, 1.0f, 1.0f); /* cyan */
    glVertex3f(-1.0f,  1.0f,  1.0f);
    glColor3f(0.0f, 1.0f, 0.0f); /* green */
    glVertex3f(-1.0f,  1.0f, -1.0f);
    glColor3f(0.0f, 0.0f, 0.0f); /* black corner */
    glVertex3f(-1.0f, -1.0f, -1.0f);

    glColor3f(1.0f, 0.0f, 0.0f); /* red */
    glVertex3f( 1.0f, -1.0f, -1.0f);
    glColor3f(1.0f, 1.0f, 0.0f);  /* yellow */
    glVertex3f( 1.0f,  1.0f, -1.0f);
    glColor3f(1.0f, 1.0f, 1.0f);  /* white */
    glVertex3f( 1.0f,  1.0f,  1.0f);
    glColor3f(1.0f, 0.0f, 1.0f);  /* purple corner */
    glVertex3f( 1.0f, -1.0f,  1.0f);
  glEnd();
}

void Draw() {
  /* Setup for the scene, Basically just clear the buffers */
  /* Wipe the color and Depth Buffers */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  /* Manipulate the PROJECTION matrix for all models
   * basically use the projection matrix to simulate observer position
   * otherwise we'd need an extra matrix transform for every point
   * this way there's just two.
   */
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();

  glRotatef(-cameraPitch, 1.0f, 0.0f, 0.0f); /* rotate about the y unit vector */
  glRotatef(-cameraYaw,   0.0f, 1.0f, 0.0f); /* rotate about the y unit vector */
  glRotatef(-cameraRoll,  0.0f, 0.0f, 1.0f); /* rotate about the y unit vector */
  glTranslatef(-cameraX, -cameraY, -cameraZ);

  /* Manipulate the MODELVIEW matrix for each model
   * the modelview matrix. Remember that you manipulate the view 
   * matrix and not the objects themselves so all so all transforms
   * are logically backwards from whats intuitive
   */
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

  /* Draw all my objects */
  DrawRGBCube();
  DrawTextureCube(texID);

  /* restore the original projection matrix */
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix(); /* restore the original model-view matrix */

  glFinish(); /* Block the program until cached rendering commands are complete */

  OpenGLSwapBuffers(hDC); /* Swap Buffers once rendering is complete */
}

void Animate(float elapsedTime) {
  texCubeYaw   += YAW_RATE*elapsedTime;
  texCubePitch += PITCH_RATE*elapsedTime;
  texCubeRoll  += ROLL_RATE*elapsedTime;

  rgbCubeYaw   -= YAW_RATE*elapsedTime;
  rgbCubePitch -= PITCH_RATE*elapsedTime;
  rgbCubeRoll  -= ROLL_RATE*elapsedTime;
}

/* 
 *
 * Standard Windows core logic
 * 
 */
int WINAPI MainWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
  LONG        lRet = 1;
  RECT        rect;
  
  switch (uMsg) {
  case WM_CREATE:
    hWndMain = hWnd;
    hDC = GetDC(hWnd);
    if (!SetupPixelFormat(hDC, 16, 16)) {
      PostQuitMessage(0);
      break;
    }
    hRC = OpenGLCreateContext(hDC);
    if (!hRC) {
      Log(LOG_ERROR, "OpenGLCreateContext failed\n");
      ShowSystemError("OpenGLCreateContext failed\n");
      PostQuitMessage(0);
      break;
    }
    OpenGLMakeCurrent(hDC, hRC);
    GetClientRect(hWnd, &rect);
    InitGL(rect.right, rect.bottom);
    srand( (unsigned)time( NULL ) );
    break;

#ifdef CUSTOM_PAINT_LOGIC    
  case WM_PAINT: {
    PAINTSTRUCT ps;

    BeginPaint(hWnd, &ps);
    /* custom paint logic would go here, if we had any */
    EndPaint(hWnd, &ps);
    break;
  }
#endif
  
  case WM_ACTIVATE: {
    static LONG oldPriority;

    if ( LOWORD(wParam)==WA_INACTIVE ) {
      OpenGLMakeCurrent(hDC, NULL); //OpenGLMakeCurrent
      ShowWindow(hWnd, SW_MINIMIZE);
      oldPriority = priority;
      priority=0;
      SetPriority(priority);
    } else {
      OpenGLMakeCurrent(hDC, hRC); //OpenGLMakeCurrent
      ShowWindow(hWnd, SW_RESTORE);
      priority=oldPriority;
      SetPriority(priority);
    }
  } break;

  case WM_SIZE:
    GetClientRect(hWnd, &rect);
    ResizeViewport(rect.right, rect.bottom);
    break;

  case WM_CLOSE:
    PostQuitMessage(0);
    lRet = 0;
    break;

  case WM_DESTROY:
    Log(LOG_BOOT, "Releasing OpenGL context\n");
    OpenGLMakeCurrent(NULL, NULL) ; 
    if (hRC) { OpenGLDeleteContext(hRC); hRC = 0; }
    if (hDC) { ReleaseDC(hWnd, hDC);  hDC = 0; }
    PostQuitMessage (0);
    break;

  default:
    lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
    break;
  }
  
  return lRet;
}


/* 
 * Core looping routine, Alt-F4 (standard windows quit key, y'know) to exit.
 */
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
  MSG       msg;
  WNDCLASS  wndclass;
  DWORD     thisTime;
  DWORD     lastTime;
  HWND      hWnd;
  HANDLE    hProc;
  GLfloat   elapsedTime;
  GLfloat   totalTime = 0.0f;
  LONG      frameNum = 0;
  LONG      frameTest = 0;
  GLfloat   testTime = 0.0f;
  LONG      priority = PRIORITY;

  hProc = GetCurrentProcess();
  
  /* Register the frame class */
  wndclass.style         = CS_OWNDC;
  wndclass.lpfnWndProc   = (WNDPROC)MainWndProc;
  wndclass.cbClsExtra    = 0;
  wndclass.cbWndExtra    = 0;
  wndclass.hInstance     = hInstance;
  wndclass.hIcon         = LoadIcon(hInstance, szAppName);
  wndclass.hCursor       = LoadCursor(NULL,IDC_ARROW);
  wndclass.hbrBackground = GetStockObject(BLACK_BRUSH);
  wndclass.lpszMenuName  = szAppName;
  wndclass.lpszClassName = szAppName;
  
  if (!RegisterClass(&wndclass))
    return FALSE;
  
  if ( FULLSCREEN ) {
    if (!ChangeResolution(WIDTH,HEIGHT,BITDEPTH)) return FALSE;
  } 

  {
    BYTE openglLib[256];
    OpenGLFindDriver(openglLib);
    if (OpenGLInit(openglLib)) {
      Log(LOG_ERROR, "Unable to load opengl dll\n");
      MessageBox(NULL, "Unable to load opengl dll", "OpenGLInit Error", MB_ICONERROR); 
      return FALSE;
    }
    Log(LOG_BOOT, "Using OpenGL Driver: %s\n", openglLib);
  }

  /* Create the frame */
  /* Use the WS_POPUP window style to bring up a window that is
  * exactly WIDTHxHEIGHT.
  */
  if ( FULLSCREEN ) {
    hWnd = CreateWindow(szAppName, "Azteroidz3D", 
                         WS_POPUP |  WS_CLIPSIBLINGS,
                        0, 0, WIDTH, HEIGHT,
                        NULL, NULL, hInstance, NULL);
  } else {
    LONG  h;
    LONG  w;

    h = GetSystemMetrics(SM_CYCAPTION) + GetSystemMetrics(SM_CXDLGFRAME)*2 + HEIGHT;
    w = GetSystemMetrics(SM_CXDLGFRAME)*2 + WIDTH;

    // Same window settings as QuakeII
    // WS_BORDER WS_DLGFRAME WS_CLIPSIBLINGS
    hWnd = CreateWindow(szAppName, "GLDemo", 
                        WS_BORDER | WS_DLGFRAME | WS_CLIPSIBLINGS,
                        1, 1, w, h,
                        NULL, NULL, hInstance, NULL);
  }


  /* Make sure window was created */
  if (!hWnd) {
    Log(LOG_ERROR, "Unable to create window\n");
    MessageBox(NULL, "Unable to create window", "CreateWindow Error", MB_ICONERROR); 
    return FALSE;
  }
  
  /* Show and update main window */
  ShowWindow(hWnd, nCmdShow);
  UpdateWindow(hWnd);

  /* update the process priority */
  SetPriority(priority);

  /* Animation loop */
  /* This ensures a fixed animation rate, frame rate will
   * vary by videocard performance
   */
  thisTime = timeGetTime();
  while (1) {
    /* Process all pending messages */
    /* If we dont lower the priority here, the keyboard / mouse events
     * wont get processed, and the computer will be hung
     */
    if (priority==3) SetPriority(2);

    while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE) == TRUE) {
      if (GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
      } else { /* WM_QUIT message */
        Log(LOG_BOOT, "Average Frames per Second %f\n", Float(frameNum) / totalTime);
        DestroyWindow(hWnd);
        Log(LOG_BOOT, "Releasing OpenGL driver\n");
        OpenGLUnInit();
        if (FULLSCREEN) {
          Log(LOG_BOOT, "Restoring desktop resolution\n");
          ChangeDisplaySettings(NULL, 0); 
        }
        return TRUE;
      }
    }

    /*restore realtime priority*/
    if (priority==3) SetPriority(3);

    /* animate the scene - Calculate Time since last frame */
    lastTime = thisTime;
    thisTime = timeGetTime();
    elapsedTime = ((float)(thisTime - lastTime) / 1000);

    frameNum += 1;
    totalTime += elapsedTime;
    if (FRAMETEST && totalTime > TESTTIME) PostQuitMessage(1);
    fps = (LONG)(1.0f / elapsedTime);
    Animate(elapsedTime);

    Draw();

  }
}


